==> libpqtypes-1.3.1.tar.gz: libpqtypes-1.3.1/src/param.c <==

/*
 * param.c
 *   Management functions for the PGparam object.
 *
 * Copyright (c) 2009 eSilo, LLC. All rights reserved.
 * This is free software; see the source for copying conditions.  There is
 * NO warranty; not even for MERCHANTABILITY or  FITNESS FOR A  PARTICULAR
 * PURPOSE.
 */

#include "libpqtypes-int.h"

static char *skipQuotes(char *s);
static int argsExpandBuffer(PGtypeArgs *args, int new_len);
static int putParam(PGparam *param, const void *data, int datal,
	int flags, int format, PGtypeHandler *handler);

PGparam *
PQparamCreate(const PGconn *conn)
{
	PGparam *param;
	PGtypeData *connData;

	PQseterror(NULL);

	if (!conn)
	{
		PQseterror("PGconn cannot be NULL");
		return NULL;
	}

	param = (PGparam *) malloc(sizeof(PGparam));
	if (!param)
	{
		PQseterror(PQT_OUTOFMEMORY);
		return NULL;
	}

	memset(param, 0, sizeof(PGparam));

	connData = (PGtypeData *) PQinstanceData(conn, pqt_eventproc);
	if (!connData)
	{
		PQparamClear(param);
		PQseterror("No type data exists for PGconn at %p", conn);
		return NULL;
	}

	/* copy any handlers from conn object */
	if (connData->typhcnt > 0)
	{
		param->typhandlers = pqt_duphandlers(
			connData->typhandlers, connData->typhcnt);

		if (!param->typhandlers)
		{
			PQparamClear(param);
			PQseterror(PQT_OUTOFMEMORY);
			return NULL;
		}

		param->typhcnt = connData->typhcnt;
	}

	pqt_getfmtinfo(conn, &param->fmtinfo);
	return param;
}

int
PQparamCount(PGparam *param)
{
	if (param)
		return param->vcnt;
	return 0;
}

void
PQparamReset(PGparam *param)
{
	if (param)
		param->vcnt = 0;
}

void
PQparamClear(PGparam *param)
{
	int i;

	if (!param)
		return;

	/* vmax is the size of the vals array.  We don't use vcnt because
	 * someone could of just called PQparamReset, leaving vcnt 0.  If
	 * we find any value pointers that are not NULL, free'm.
	 */
	for (i=0; i < param->vmax; i++)
		if (param->vals[i].ptr)
			free(param->vals[i].ptr);

	if (param->vals)
		free(param->vals);

	pqt_freehandlers(param->typhandlers, param->typhcnt);
	pqt_freespeccache(&param->cache);

	param->vmax = 0;
	param->vcnt = 0;
	param->vals = NULL;
	param->typhcnt = 0;
	param->typhandlers = NULL;

	free(param);
}

int
PQputf(PGparam *param, const char *spec, ...)
{
	int r;
	va_list ap;

	PQseterror(NULL);

	if (!param)
	{
		PQseterror("PGparam cannot be NULL");
		return FALSE;
	}

	if (!spec || !*spec)
	{
		PQseterror("spec cannot be NULL or an empty string");
		return FALSE;
	}

	/* This function is just a wrapper to PGputvf */
	va_start(ap, spec);
	r = PQputvf(param, NULL, 0, spec, ap);
	va_end(ap);

	return r;
}

/* For use with PQputvf only */
#define CHKSTMTBUF(nbytes_add) do{ \
	if ((stmtPos + (nbytes_add)) >= stmtBufLen) \
	{ \
		PQseterror("statement buffer is too small"); \
		param->vcnt = save_vcnt; \
		return FALSE; \
	} \
}while (0)

int
PQputvf(PGparam *param, char *stmtBuf, size_t stmtBufLen,
	const char *spec, va_list ap)
{
	int n=0;
	int flags;
	char *s;
	size_t stmtPos = 0;
	int save_vcnt;
	int typpos = 0;
	PGtypeHandler *h;
	PGtypeArgs args;
	char args_outbuf[4096];
	char schema[PQT_MAXIDLEN + 1];
	char typname[PQT_MAXIDLEN + 1];
	char tmp[(PQT_MAXIDLEN*2)+4];
	int have_cache = 0;
	int idmax = 0;
	const char *save_spec = spec;

	PQseterror(NULL);

	if (!param)
	{
		PQseterror("PGparam cannot be NULL");
		return FALSE;
	}

	if (!spec || !*spec)
	{
		PQseterror("param 'spec' cannot be NULL or an empty string");
		return FALSE;
	}

	if (stmtBuf && stmtBufLen < 1)
	{
		PQseterror("Invalid argument: stmtBufLen must be >= 1");
		return FALSE;
	}

	save_vcnt = param->vcnt;
	va_copy(args.ap, ap);

	if (param->cache.spec && param->cache.idcnt > 0)
		have_cache = strcmp(param->cache.spec, spec)==0 ? 1 : 0;

	if (!have_cache)
		pqt_freespeccache(&param->cache);

	while (spec && *spec)
	{
		if (have_cache)
		{
			/* done, no more handlers in cached spec string. */
			if (typpos == param->cache.idcnt)
				break;

			h = pqt_gethandlerbyid(param->typhandlers, param->typhcnt,
				param->cache.idlist[typpos]);

			/* should be an unusual, or a "will never happen", situation */
			if (!h)
			{
				va_end(args.ap);
				PQseterror("Uknown type handler id at position %d", typpos+1);
				pqt_freespeccache(&param->cache);
				param->vcnt = save_vcnt;
				return FALSE;
			}

			flags = (int)param->cache.flags[typpos];
			typpos++;
		}
		else
		{
			if (!(s = skipQuotes((char *) spec)))
			{
				param->vcnt = save_vcnt;
				return FALSE;
			}

			/* found quotes to skip */
			if (s != spec)
			{
				if (stmtBuf)
				{
					CHKSTMTBUF(s-spec);
					memcpy(stmtBuf + stmtPos, spec, s-spec);
					stmtPos += (s-spec);
				}

				spec = s;
				continue;
			}

			if (*spec != '%' && *spec != '#')
			{
				if (stmtBuf)
				{
					CHKSTMTBUF(1);
					stmtBuf[stmtPos++] = *spec;
				}

				spec++;
				continue;
			}

			/* spec skips % or # */
			if (!(s = pqt_parsetype(spec + 1, schema, typname, &flags, typpos + 1)))
			{
				param->vcnt = save_vcnt;
				return FALSE;
			}

			if (flags & TYPFLAG_INVALID)
			{
				if (stmtBuf)
				{
					CHKSTMTBUF(1);
					stmtBuf[stmtPos++] = *spec++;
					PQseterror(NULL); /* set by pqt_parsetype */
					continue;
				}

				/* Error, PQputf doesn't allow arbitrary % or # in spec string */
				param->vcnt = save_vcnt;
				return FALSE;
			}

			typpos++;

			if (!(h = pqt_gethandler(param->typhandlers, param->typhcnt,
				schema, typname)))
			{
				param->vcnt = save_vcnt;
				PQseterror("Uknown type '%s' (position %d)",
					pqt_fqtn(tmp, sizeof(tmp), schema, typname), typpos);
				return FALSE;
			}

			spec = s;
			if (stmtBuf)
			{
				n = pqt_snprintf(tmp, sizeof(tmp), "$%d", typpos);
				CHKSTMTBUF(n);
				memcpy(stmtBuf + stmtPos, tmp, n);
				stmtPos += n;
			}

			if (!h->typput)
			{
				h = pqt_gethandlerbyid(param->typhandlers, param->typhcnt, h->base_id);
				if (!h || !h->typput)
				{
					param->vcnt = save_vcnt;
					PQseterror(
						"Type '%s' doesn't support put operations (position %d)",
						pqt_fqtn(tmp, sizeof(tmp), h->typschema, h->typname), typpos);
					return FALSE;
				}
			}

			if ((flags & TYPFLAG_POINTER) && !pqt_allowsptr(h))
			{
				param->vcnt = save_vcnt;
				PQseterror(
					"Type '%s' doesn't support putting pointers (position %d)",
					pqt_fqtn(tmp, sizeof(tmp), h->typschema, h->typname), typpos);
				return FALSE;
			}
		}

		args.is_put           = 1;
		args.put.param        = param;
		args.fmtinfo          = &param->fmtinfo;
		args.put.out          = args_outbuf;
		args.put.__allocated_out = NULL;
		args.put.outl         = (int) sizeof(args_outbuf);
		args.is_ptr           = (flags & TYPFLAG_POINTER) ? 1 : 0;
		args.format           = BINARYFMT;
		args.put.expandBuffer = argsExpandBuffer;
		args.typpos           = typpos;
		args.typhandler       = h;
		args.errorf           = pqt_argserrorf;
		args.super            = pqt_argssuper;
		*args.put.out         = 0;

		if (flags & TYPFLAG_ARRAY)
			n = pqt_put_array(&args);
		else
			n = h->typput(&args);

		if (n == -1)
		{
			if (args.put.__allocated_out && args.put.__allocated_out != args_outbuf)
				free(args.put.__allocated_out);
			param->vcnt = save_vcnt;
			return FALSE;
		}

		if (args.put.out == NULL)
		{
			args.format = BINARYFMT;
			n = -1;
		}

		n = putParam(param, args.put.out, n, flags, args.format, h);

		if (args.put.__allocated_out && args.put.__allocated_out != args_outbuf)
			free(args.put.__allocated_out);

		if (!n)
		{
			param->vcnt = save_vcnt;
			return FALSE;
		}

		if (!have_cache)
		{
			if ((idmax = pqt_growspeccache(&param->cache, idmax)) == -1)
			{
				va_end(args.ap);
				pqt_freespeccache(&param->cache);
				param->vcnt = save_vcnt;
				return FALSE;
			}

			/* Parallel arrays, every handler needs type flags */
			param->cache.idlist[param->cache.idcnt] = h->id;
			param->cache.flags[param->cache.idcnt++] = (unsigned char) flags;
		}
	}

	/* If no cache yet, we have to terminate stmtBuf and copy the field
	 * spec string and stmtBuf to the param's spec-cache.
	 */
	if (!have_cache)
	{
		/* terminate stmtBuf, guarenteed to have room for NUL */
		if (stmtBuf)
			stmtBuf[stmtPos] = 0;

		/* dup spec string */
		param->cache.spec = strdup(save_spec);

		/* ignore error since we processed all fields already.
		 * Although, we must clear cache.
		 */
		if (!param->cache.spec)
		{
			pqt_freespeccache(&param->cache);
		}
		else if (stmtBuf)
		{
			/* cache the parameterized stmt string */
			param->cache.stmt = strdup(stmtBuf);
			if (!param->cache.stmt)
				pqt_freespeccache(&param->cache);
		}
	}
	/* Have a cache, copy the param's cached stmt to stmtBuf */
	else if (stmtBuf)
	{
		size_t n = strlen(param->cache.stmt);
		CHKSTMTBUF(n);
		memcpy(stmtBuf, param->cache.stmt, n + 1);
	}

	return TRUE;
}


/* ----------------------------
 * Helper functions
 */

static int
putParam(PGparam *param, const void *data, int datal, int flags,
	int format, PGtypeHandler *h)
{
	PGvalue *v;

	if (!param || !h)
		return FALSE;

	if (!data)
		datal = -1;

	/* need to grow param vals array */
	if (param->vcnt == param->vmax)
	{
		PGvalue *vals;
		int vmax = param->vmax ? (param->vmax * 3) / 2 : 16;

		vals = (PGvalue *) pqt_realloc(param->vals, sizeof(PGvalue) * vmax);
		if (!vals)
		{
			PQseterror(PQT_OUTOFMEMORY);
			return FALSE;
		}

		/* zero out the new array elements */
		memset(vals + param->vcnt, 0, (vmax - param->vcnt) * sizeof(PGvalue));
		param->vmax = vmax;
		param->vals = vals;
	}

	/* grab next param value */
	v = &param->vals[param->vcnt];

	if (datal == -1)
	{
		v->data = NULL;
	}
	/* wants to put a direct pointer */
	else if (flags & TYPFLAG_POINTER)
	{
		v->data = (char *) data;
	}
	else
	{
		/* need more mem for param value ptr */
		if (v->ptrl < datal)
		{
			char *ptr = (char *) pqt_realloc(v->ptr, datal);

			if (!ptr)
			{
				PQseterror(PQT_OUTOFMEMORY);
				return FALSE;
			}

			v->ptrl = datal;
			v->ptr = ptr;
		}

		memcpy(v->ptr, data, datal);
		v->data = v->ptr;
	}

	v->datal	= datal;
	v->format = format;
	v->oid		= (flags & TYPFLAG_ARRAY) ? h->typoid_array : h->typoid;
	param->vcnt++;
	return TRUE;
}

/* returns 0 on success and -1 on error */
static int
argsExpandBuffer(PGtypeArgs *args, int new_len)
{
	char *new_out;

	if (new_len <= args->put.outl)
		return 0;

	if (!args->put.__allocated_out)
	{
		new_out = (char *) malloc(new_len);
		if (!new_out)
			return args->errorf(args, PQT_OUTOFMEMORY);

		/* fully copy existing stack buffer, we don't know what data
		 * the user has already written and may want to keep.
		 */
		memcpy(new_out, args->put.out, args->put.outl);
	}
	else
	{
		new_out = (char *) realloc(args->put.__allocated_out, new_len);
		if (!new_out)
			return args->errorf(args, PQT_OUTOFMEMORY);
	}

	args->put.out = args->put.__allocated_out = new_out;
	args->put.outl = new_len;
	return 0;
}

/* skip quoted strings.  Doesn't need to account for E'' syntax. The
 * E is copied over prior to the quoted string.
 *
 * Returns a pointer to the next character after the closing quote or
 * NULL if there was an error.
 */
static char *
skipQuotes(char *s)
{
	char *end;

	if (*s != '\'')
		return s;

	end = s;
	while (*++end)
	{
		/* If we see a backslash, skip an extra char.  No need to dig any
		 * further since this method works with \digits and \hex.
		 */
		if (*end == '\\')
			end++;
		else if (*end == '\'')
			break;
	}

	/* unterminated quote */
	if (!*end)
	{
		PQseterror("unterminated single quoted string");
		return NULL;
	}

	return ++end; /* skip ending quote */
}

==> libpqtypes-1.3.2.tar.gz: libpqtypes-1.3.2/src/param.c <==

/*
 * param.c
 *   Management functions for the PGparam object.
 *
 * Copyright (c) 2009 eSilo, LLC. All rights reserved.
 * This is free software; see the source for copying conditions.  There is
 * NO warranty; not even for MERCHANTABILITY or  FITNESS FOR A  PARTICULAR
 * PURPOSE.
 */

#include "libpqtypes-int.h"

static char *skipQuotes(char *s);
static int argsExpandBuffer(PGtypeArgs *args, int new_len);

PGparam *
PQparamCreate(const PGconn *conn)
{
	PGparam *param;
	PGtypeData *connData;

	PQseterror(NULL);

	if (!conn)
	{
		PQseterror("PGconn cannot be NULL");
		return NULL;
	}

	param = (PGparam *) malloc(sizeof(PGparam));
	if (!param)
	{
		PQseterror(PQT_OUTOFMEMORY);
		return NULL;
	}

	memset(param, 0, sizeof(PGparam));

	connData = (PGtypeData *) PQinstanceData(conn, pqt_eventproc);
	if (!connData)
	{
		PQparamClear(param);
		PQseterror("No type data exists for PGconn at %p", conn);
		return NULL;
	}

	/* copy any handlers from conn object */
	if (connData->typhcnt > 0)
	{
		param->typhandlers = pqt_duphandlers(
			connData->typhandlers, connData->typhcnt);

		if (!param->typhandlers)
		{
			PQparamClear(param);
			PQseterror(PQT_OUTOFMEMORY);
			return NULL;
		}

		param->typhcnt = connData->typhcnt;
	}

	pqt_getfmtinfo(conn, &param->fmtinfo);
	return param;
}

int
PQparamCount(PGparam *param)
{
	if (param)
		return param->vcnt;
	return 0;
}

void
PQparamReset(PGparam *param)
{
	if (param)
		param->vcnt = 0;
}

void
PQparamClear(PGparam *param)
{
	int i;

	if (!param)
		return;

	/* vmax is the size of the vals array.  We don't use vcnt because
	 * someone could of just called PQparamReset, leaving vcnt 0.  If
	 * we find any value pointers that are not NULL, free'm.
	 */
	for (i=0; i < param->vmax; i++)
		if (param->vals[i].ptr)
			free(param->vals[i].ptr);

	if (param->vals)
		free(param->vals);

	pqt_freehandlers(param->typhandlers, param->typhcnt);
	pqt_freespeccache(&param->cache);

	param->vmax = 0;
	param->vcnt = 0;
	param->vals = NULL;
	param->typhcnt = 0;
	param->typhandlers = NULL;

	free(param);
}

int
PQputf(PGparam *param, const char *spec, ...)
{
	int r;
	va_list ap;

	PQseterror(NULL);

	if (!param)
	{
		PQseterror("PGparam cannot be NULL");
		return FALSE;
	}

	if (!spec || !*spec)
	{
		PQseterror("spec cannot be NULL or an empty string");
		return FALSE;
	}

	/* This function is just a wrapper to PGputvf */
	va_start(ap, spec);
	r = PQputvf(param, NULL, 0, spec, ap);
	va_end(ap);

	return r;
}

/* For use with PQputvf only */
#define CHKSTMTBUF(nbytes_add) do{ \
	if ((stmtPos + (nbytes_add)) >= stmtBufLen) \
	{ \
		PQseterror("statement buffer is too small"); \
		param->vcnt = save_vcnt; \
		return FALSE; \
	} \
}while (0)

int
PQputvf(PGparam *param, char *stmtBuf, size_t stmtBufLen,
	const char *spec, va_list ap)
{
	int n=0;
	int flags;
	char *s;
	size_t stmtPos = 0;
	int save_vcnt;
	int typpos = 0;
	PGtypeHandler *h;
	PGtypeArgs args;
	char args_outbuf[4096];
	char schema[PQT_MAXIDLEN + 1];
	char typname[PQT_MAXIDLEN + 1];
	char tmp[(PQT_MAXIDLEN*2)+4];
	int have_cache = 0;
	int idmax = 0;
	const char *save_spec = spec;

	PQseterror(NULL);

	if (!param)
	{
		PQseterror("PGparam cannot be NULL");
		return FALSE;
	}

	if (!spec || !*spec)
	{
		PQseterror("param 'spec' cannot be NULL or an empty string");
		return FALSE;
	}

	if (stmtBuf && stmtBufLen < 1)
	{
		PQseterror("Invalid argument: stmtBufLen must be >= 1");
		return FALSE;
	}

	save_vcnt = param->vcnt;
	va_copy(args.ap, ap);

	if (param->cache.spec && param->cache.idcnt > 0)
		have_cache = strcmp(param->cache.spec, spec)==0 ? 1 : 0;

	if (!have_cache)
		pqt_freespeccache(&param->cache);

	while (spec && *spec)
	{
		if (have_cache)
		{
			/* done, no more handlers in cached spec string. */
			if (typpos == param->cache.idcnt)
				break;

			h = pqt_gethandlerbyid(param->typhandlers, param->typhcnt,
				param->cache.idlist[typpos]);

			/* should be an unusual, or a "will never happen", situation */
			if (!h)
			{
				va_end(args.ap);
				PQseterror("Uknown type handler id at position %d", typpos+1);
				pqt_freespeccache(&param->cache);
				param->vcnt = save_vcnt;
				return FALSE;
			}

			flags = (int)param->cache.flags[typpos];
			typpos++;
		}
		else
		{
			if (!(s = skipQuotes((char *) spec)))
			{
				param->vcnt = save_vcnt;
				return FALSE;
			}

			/* found quotes to skip */
			if (s != spec)
			{
				if (stmtBuf)
				{
					CHKSTMTBUF(s-spec);
					memcpy(stmtBuf + stmtPos, spec, s-spec);
					stmtPos += (s-spec);
				}

				spec = s;
				continue;
			}

			if (*spec != '%' && *spec != '#')
			{
				if (stmtBuf)
				{
					CHKSTMTBUF(1);
					stmtBuf[stmtPos++] = *spec;
				}

				spec++;
				continue;
			}

			/* spec skips % or # */
			if (!(s = pqt_parsetype(spec + 1, schema, typname, &flags, typpos + 1)))
			{
				param->vcnt = save_vcnt;
				return FALSE;
			}

			if (flags & TYPFLAG_INVALID)
			{
				if (stmtBuf)
				{
					CHKSTMTBUF(1);
					stmtBuf[stmtPos++] = *spec++;
					PQseterror(NULL); /* set by pqt_parsetype */
					continue;
				}

				/* Error, PQputf doesn't allow arbitrary % or # in spec string */
				param->vcnt = save_vcnt;
				return FALSE;
			}

			typpos++;

			if (!(h = pqt_gethandler(param->typhandlers, param->typhcnt,
				schema, typname)))
			{
				param->vcnt = save_vcnt;
				PQseterror("Uknown type '%s' (position %d)",
					pqt_fqtn(tmp, sizeof(tmp), schema, typname), typpos);
				return FALSE;
			}

			spec = s;
			if (stmtBuf)
			{
				n = pqt_snprintf(tmp, sizeof(tmp), "$%d", typpos);
				CHKSTMTBUF(n);
				memcpy(stmtBuf + stmtPos, tmp, n);
				stmtPos += n;
			}

			if (!h->typput)
			{
				h = pqt_gethandlerbyid(param->typhandlers, param->typhcnt, h->base_id);
				if (!h || !h->typput)
				{
					param->vcnt = save_vcnt;
					PQseterror(
						"Type '%s' doesn't support put operations (position %d)",
						pqt_fqtn(tmp, sizeof(tmp), h->typschema, h->typname), typpos);
					return FALSE;
				}
			}

			if ((flags & TYPFLAG_POINTER) && !pqt_allowsptr(h))
			{
				param->vcnt = save_vcnt;
				PQseterror(
					"Type '%s' doesn't support putting pointers (position %d)",
					pqt_fqtn(tmp, sizeof(tmp), h->typschema, h->typname), typpos);
				return FALSE;
			}
		}

		args.is_put           = 1;
		args.put.param        = param;
		args.fmtinfo          = &param->fmtinfo;
		args.put.out          = args_outbuf;
		args.put.__allocated_out = NULL;
		args.put.outl         = (int) sizeof(args_outbuf);
		args.is_ptr           = (flags & TYPFLAG_POINTER) ? 1 : 0;
		args.format           = BINARYFMT;
		args.put.expandBuffer = argsExpandBuffer;
		args.typpos           = typpos;
		args.typhandler       = h;
		args.errorf           = pqt_argserrorf;
		args.super            = pqt_argssuper;
		*args.put.out         = 0;

		if (flags & TYPFLAG_ARRAY)
			n = pqt_put_array(&args);
		else
			n = h->typput(&args);

		if (n == -1)
		{
			if (args.put.__allocated_out && args.put.__allocated_out != args_outbuf)
				free(args.put.__allocated_out);
			param->vcnt = save_vcnt;
			return FALSE;
		}

		if (args.put.out == NULL)
		{
			args.format = BINARYFMT;
			n = -1;
		}

		n = pqt_putparam(param, args.put.out, n, flags, args.format,
			(flags & TYPFLAG_ARRAY) ? h->typoid_array : h->typoid);

		if (args.put.__allocated_out && args.put.__allocated_out != args_outbuf)
			free(args.put.__allocated_out);

		if (!n)
		{
			param->vcnt = save_vcnt;
			return FALSE;
		}

		if (!have_cache)
		{
			if ((idmax = pqt_growspeccache(&param->cache, idmax)) == -1)
			{
				va_end(args.ap);
				pqt_freespeccache(&param->cache);
				param->vcnt = save_vcnt;
				return FALSE;
			}

			/* Parallel arrays, every handler needs type flags */
			param->cache.idlist[param->cache.idcnt] = h->id;
			param->cache.flags[param->cache.idcnt++] = (unsigned char) flags;
		}
	}

	/* If no cache yet, we have to terminate stmtBuf and copy the field
	 * spec string and stmtBuf to the param's spec-cache.
	 */
	if (!have_cache)
	{
		/* terminate stmtBuf, guarenteed to have room for NUL */
		if (stmtBuf)
			stmtBuf[stmtPos] = 0;

		/* dup spec string */
		param->cache.spec = strdup(save_spec);

		/* ignore error since we processed all fields already.
		 * Although, we must clear cache.
		 */
		if (!param->cache.spec)
		{
			pqt_freespeccache(&param->cache);
		}
		else if (stmtBuf)
		{
			/* cache the parameterized stmt string */
			param->cache.stmt = strdup(stmtBuf);
			if (!param->cache.stmt)
				pqt_freespeccache(&param->cache);
		}
	}
	/* Have a cache, copy the param's cached stmt to stmtBuf */
	else if (stmtBuf)
	{
		size_t n = strlen(param->cache.stmt);
		CHKSTMTBUF(n);
		memcpy(stmtBuf, param->cache.stmt, n + 1);
	}

	return TRUE;
}


/* ----------------------------
 * Helper functions
 */

int
pqt_putparam(PGparam *param, const void *data, int datal,
	int flags, int format, Oid typoid)
{
	PGvalue *v;

	if (!param)
		return FALSE;

	if (!data)
		datal = -1;

	/* need to grow param vals array */
	if (param->vcnt == param->vmax)
	{
		PGvalue *vals;
		int vmax = param->vmax ? (param->vmax * 3) / 2 : 16;

		vals = (PGvalue *) pqt_realloc(param->vals, sizeof(PGvalue) * vmax);
		if (!vals)
		{
			PQseterror(PQT_OUTOFMEMORY);
			return FALSE;
		}

		/* zero out the new array elements */
		memset(vals + param->vcnt, 0, (vmax - param->vcnt) * sizeof(PGvalue));
		param->vmax = vmax;
		param->vals = vals;
	}

	/* grab next param value */
	v = &param->vals[param->vcnt];

	if (datal == -1)
	{
		v->data = NULL;
	}
	/* wants to put a direct pointer */
	else if (flags & TYPFLAG_POINTER)
	{
		v->data = (char *) data;
	}
	else
	{
		/* need more mem for param value ptr */
		if (v->ptrl < datal)
		{
			char *ptr = (char *) pqt_realloc(v->ptr, datal);

			if (!ptr)
			{
				PQseterror(PQT_OUTOFMEMORY);
				return FALSE;
			}

			v->ptrl = datal;
			v->ptr = ptr;
		}

		memcpy(v->ptr, data, datal);
		v->data = v->ptr;
	}

	v->datal	= datal;
	v->format = format;
	v->oid		= typoid;
	param->vcnt++;
	return TRUE;
}

/* returns 0 on success and -1 on error */
static int
argsExpandBuffer(PGtypeArgs *args, int new_len)
{
	char *new_out;

	if (new_len <= args->put.outl)
		return 0;

	if (!args->put.__allocated_out)
	{
		new_out = (char *) malloc(new_len);
		if (!new_out)
			return args->errorf(args, PQT_OUTOFMEMORY);

		/* fully copy existing stack buffer, we don't know what data
		 * the user has already written and may want to keep.
		 */
		memcpy(new_out, args->put.out, args->put.outl);
	}
	else
	{
		new_out = (char *) realloc(args->put.__allocated_out, new_len);
		if (!new_out)
			return args->errorf(args, PQT_OUTOFMEMORY);
	}

	args->put.out = args->put.__allocated_out = new_out;
	args->put.outl = new_len;
	return 0;
}

/* skip quoted strings.  Doesn't need to account for E'' syntax. The
 * E is copied over prior to the quoted string.
 *
 * Returns a pointer to the next character after the closing quote or
 * NULL if there was an error.
 */
static char *
skipQuotes(char *s)
{
	char *end;

	if (*s != '\'')
		return s;

	end = s;
	while (*++end)
	{
		/* If we see a backslash, skip an extra char.  No need to dig any
		 * further since this method works with \digits and \hex.
		 */
		if (*end == '\\')
			end++;
		else if (*end == '\'')
			break;
	}

	/* unterminated quote */
	if (!*end)
	{
		PQseterror("unterminated single quoted string");
		return NULL;
	}

	return ++end; /* skip ending quote */
}
